__author__ = 'maximmillen'

from collections import OrderedDict
from sortedcontainers import SortedDict
import numpy as np

from single import Record

class Cluster(object):

    def __init__(self, master_motion, master_step, slave_motions=[], slave_steps=[], slave_names=[], **kwargs):

        self.freq_range = np.array(kwargs.get('freq_range', [0.1, 20]))
        lvt = np.log10(1.0 / self.freq_range)
        self.response_times = kwargs.get('resp_times', np.logspace(lvt[1], lvt[0], 31, base=10))

        self.master = Record(master_motion, master_step, response_times=self.response_times)
        if len(slave_motions) > slave_names:
            shortage = len(slave_motions) - slave_names
            for i in range(shortage):
                slave_names.append("m%i" % (len(slave_names) + i))
        self.slaves = SortedDict()
        for s in range(len(slave_motions)):
            self.slaves[slave_names[s]] = Record(slave_motions[s], slave_steps[s],
                                                 slave_names[s], response_times=self.response_times)



    def motions(self, index):
        if index == 0:
            return self.master
        return self.slaves.iloc[index - 1]

    def same_start(self, **kwargs):
        '''
        Starts both time series at the same y-value
        '''
        base = kwargs.get('base', 0)
        start = kwargs.get('start', 0)
        end = kwargs.get('end', 1)
        verbose = kwargs.get('verbose', 0)
        master_average = self.motions(0).get_section_average(start=start, end=end)
        slave_average = self.motions(1).get_section_average(start=start, end=end)
        diff = slave_average - master_average
        if verbose:
            print('Same start the records')
            print('old difference in starts: ', diff)

        slave = self.motions(1)
        slave.motion -= diff

    def combine_motions(self, f_ch, **kwargs):
        '''
        This method combined the two ground motions by taking the high frequency of one values
        and the low frequency of the other.
        WARNING: records must have the same time step!!!
        f_ch: is the frequency change point
        order: refers to the order of the BW filter
        highfreq: if 0 then values[0] is used for high frequency content
                  if 1 then values[1] is used for high frequency content
        combined values is returned as self.combo
        '''
        order = kwargs.get('order', 4)
        highfreq = kwargs.get('highfreq', 0)
        remove_Gibbs = kwargs.get('remove_Gibbs', 0)
        self.master.butter_pass(f_ch, single=highfreq, filter_type='highpass', order=order, remove_Gibbs=remove_Gibbs)
        self.motions(1).butter_pass(f_ch, single=np.mod(highfreq + 1, 2), filter_type='lowpass', order=order, remove_Gibbs=remove_Gibbs)
        motion = self.motions(0).motion + self.motions(1).motion
        return motion

    def time_match(self, **kwargs):
        '''
        This method determines the time lag between two records and removes the lag
        - the solution is based on a sum of the squared residuals.
        - The motions are both set to the base values, default = 0
        - must have same time step and same magnitude
        '''
        base = kwargs.get('base', 0)
        verbose = kwargs.get('verbose', 0)
        steps = kwargs.get('steps', 10)
        set_step = kwargs.get('set_step', False)
        trim = kwargs.get('trim', True)

        if set_step == False:
            if verbose:
                print('length m0: ', self.motions(0).npts)
                print('length m1: ', self.motions(1).npts)
            length_check = min(self.motions(0).npts, self.motions(1).npts)
            bm = self.motions([np.mod(base, 2)]).motion[:length_check]
            om = self.motions([np.mod(base + 1, 2)]).motion[:length_check]
            squares = (bm[0:-steps] - om[0:-steps]) ** 2
            min_diff = np.sum(squares)
            min_ind = 0
            if verbose:
                    print('mi: ', min_ind, ' min_diff: ', min_diff)
            # Check other values lags base values
            for i in range(steps):
                squares = (om[i:-steps + i] - bm[0:-steps]) ** 2

                diff = sum(squares)
                if verbose:
                    print('ind: ', i, ' diff: ', diff)
                if diff < min_diff:
                    min_diff = diff
                    min_ind = i + 0
            # Check base values lags other values
            for i in range(steps):
                squares = (bm[i:-steps + i] - om[0:-steps]) ** 2
                diff = sum(squares)
                if verbose:
                    print('ind: ', i, ' diff: ', diff)
                if diff < min_diff:
                    min_diff = diff
                    min_ind = -i - 0

        else:
            min_ind = set_step
        if verbose:
            print('time step lag: ', min_ind)
        if min_ind == 0:
            if verbose:
                print('no time lag')
        elif min_ind < 0:
            m_temp = self.motions([np.mod(base + 1, 2)]).motion[:min_ind]
        elif min_ind > 0:
            m_temp = self.motions([np.mod(base + 1, 2)]).motion[min_ind:]
        # pad other record with constant
        if verbose:
            print('m_temp[0]: ', m_temp[0])
        if min_ind != 0:
            m = self.motions([np.mod(base + 1, 2)])
            m = np.ones(len(self.motions[np.mod(base + 1, 2)]))
        if min_ind < 0:
            self.motions[np.mod(base + 1, 2)] = m_temp[0] * self.motions[np.mod(base + 1, 2)]
            self.motions[np.mod(base + 1, 2)][-min_ind:] = m_temp
        elif min_ind > 0:
            self.motions[np.mod(base + 1, 2)] = m_temp[-1] * self.motions[np.mod(base + 1, 2)]
            self.motions[np.mod(base + 1, 2)][:-min_ind] = m_temp
        if trim:
            tmax = len(self.motions[base]) * self.dts[base]
            ender = int(tmax / self.dts[np.mod(base + 1, 2)])
            if verbose:
                print('original length: ', len(self.motions[np.mod(base + 1, 2)]))
                print('trim length: ', ender)
            self.motions[np.mod(base + 1, 2)] = self.motions[np.mod(base + 1, 2)][:ender]
        return min_ind

    def generate_response_spectrums(self):
        for i in range(len(self.slaves) + 1):
            self.motions(i).generate_response_spectrum()

    def calculate_ratios(self):

        self.generate_response_spectrums()

        self.resp_log_ratio = np.log10(self.motions(1).s_a / self.motions(0).s_a)
        self.fa_log_ratio = np.log10(self.motions(1).smooth_fa_spectrum / self.motions(0).smooth_fa_spectrum)
        centre_freq = 1.0 / self.response_times[len(self.response_times) / 2]

        self.response_ratio_moment = sum(self.resp_log_ratio * (np.log10((1.0 /
                                         self.response_times) / centre_freq)))


__author__ = 'maximmillen'

import numpy as np
import scipy.signal as ss
import scipy

from eqsig import exceptions
import eqsig.duhamels as dh
import eqsig.displacements as sd
import eqsig.measures as sm


class Signal(object):
    _npts = None
    _smooth_freq_points = 61
    _fa_spectrum = None
    _fa_frequencies = None
    _cached_fa = False
    _cached_smooth_fa = False

    def __init__(self, values, dt, label='m1', smooth_freq_range=(0.1, 30), verbose=0, ccbox=0):
        """
        A time series object

        :param values: values of the time series
        :param dt: the time step
        :param label: A name for the signal
        :param smooth_freq_range: the frequency range for plotting
        :param verbose: level of console output
        :param ccbox: colour id
        :return:
        """
        self.verbose = verbose
        self._dt = dt
        self._values = np.array(values)  # TODO: auto computes all properties (maybe just in debugger), make them lazy
        self.label = label
        self._smooth_freq_range = smooth_freq_range
        self._npts = len(self.values)
        self._smooth_fa_spectrum = np.zeros((self.smooth_freq_points))
        self._smooth_fa_frequencies = np.zeros(self.smooth_freq_points)
        self.ccbox = ccbox

    @property
    def values(self):
        return self._values

    @property
    def dt(self):
        """ The time step """
        return self._dt

    @property
    def npts(self):  # Deliberately no public setter method for this
        """ The number of points in the time series """
        return self._npts

    @property
    def time(self):
        """ An array of time of equal length to the time series """
        return np.arange(0, self.npts) * self.dt

    @property
    def fa_spectrum(self):
        """
        Generate the one-sided Fourier Amplitude spectrum
        """
        if not self._cached_fa:
            self.generate_fa_spectrum()
        return self._fa_spectrum

    def generate_fa_spectrum(self):
        n_factor = 2 ** int(np.ceil(np.log2(self.npts)))
        fa = scipy.fft(self.values, n=n_factor)
        points = int(n_factor / 2)
        self._fa_spectrum = fa[range(points)] * self.dt
        self._fa_frequencies = np.arange(points) / (2 * points * self.dt)
        self._cached_fa = True

    @property
    def fa_frequencies(self):
        if not self._cached_fa:
            self.generate_fa_spectrum()
        return self._fa_frequencies

    @property
    def smooth_freq_range(self):
        return self._smooth_freq_range

    @smooth_freq_range.setter
    def smooth_freq_range(self, limits):
        self._smooth_freq_range = np.array(limits)
        self._cached_smooth_fa = False

    @property
    def smooth_freq_points(self):
        return self._smooth_freq_points

    @smooth_freq_points.setter
    def smooth_freq_points(self, value):
        self._smooth_freq_points = value
        self._cached_smooth_fa = False

    @property
    def smooth_fa_frequencies(self):
        if not self._cached_smooth_fa:
            self.generate_smooth_fa_spectrum()
        return self._smooth_fa_frequencies

    @property
    def smooth_fa_spectrum(self):
        if not self._cached_smooth_fa:
            self.generate_smooth_fa_spectrum()
        return self._smooth_fa_spectrum

    def clear_cache(self):
        """
        Resets the dynamically calculated properties.
        :return:
        """
        self._cached_smooth_fa = False
        self._cached_fa = False

    def generate_smooth_fa_spectrum(self, band=40):
        """
        Calculates the smoothed Fourier Amplitude Spectrum
        using the method by Konno and Ohmachi 1998

        :param band: range to smooth over
        """
        lf = np.log10(self.smooth_freq_range)
        smooth_fa_frequencies = np.logspace(lf[0], lf[1], self.smooth_freq_points, base=10)
        fa_frequencies = self.fa_frequencies
        fa_spectrum = self.fa_spectrum
        for i in range(smooth_fa_frequencies.size):
            f_centre = smooth_fa_frequencies[i]
            amp_array = np.log10((fa_frequencies / f_centre) ** band)

            amp_array[0] = 0

            wb_vals = np.zeros((len(amp_array)))
            for j in range(len(amp_array)):
                if amp_array[j] == 0:
                    wb_vals[j] = 1
                else:
                    wb_vals[j] = (np.sin(amp_array[j]) / amp_array[j]) ** 4

            self._smooth_fa_spectrum[i] = (sum(abs(fa_spectrum) * wb_vals) / sum(wb_vals))
        self._smooth_fa_frequencies = smooth_fa_frequencies
        self._cached_smooth_fa = True

    def butter_pass(self, cut_off=(0.1, 15), **kwargs):
        """
        Performs a Butterworth filter

        Notes
        -----
        Wraps the scipy 'butter' filter

        Parameters
        ----------
        :param cut_off: Tuple, The cut off frequencies for the filter
        :param kwargs:
        :return:
        """
        if isinstance(cut_off, list) or isinstance(cut_off, tuple) or isinstance(cut_off, np.array):
            pass
        else:
            raise ValueError("cut_off must be list, tuple or array.")
        if len(cut_off) != 2:
            raise ValueError("cut_off must be length 2.")
        if cut_off[0] is not None and cut_off[1] is not None:
            filter_type = "band"
            cut_off = np.array(cut_off)
        elif cut_off[0] is None:
            filter_type = 'low'
            cut_off = cut_off[1]
        else:
            filter_type = 'high'
            cut_off = cut_off[0]

        filter_order = kwargs.get('filter_order', 4)
        remove_gibbs = kwargs.get('remove_gibbs', 0)
        gibbs_extra = kwargs.get('gibbs_extra', 1)
        gibbs_range = kwargs.get('gibbs_range', 50)
        sampling_rate = 1.0 / self.dt
        nyq = sampling_rate * 0.5

        mote = self.values
        org_len = len(mote)

        if remove_gibbs:
            # Pad end of record with extra zeros then cut it off after filtering
            nindex = int(np.ceil(np.log2(len(mote)))) + gibbs_extra
            new_len = 2 ** nindex
            diff_len = new_len - org_len
            if remove_gibbs == 'start':
                s_len = 0
                f_len = s_len + org_len
            elif remove_gibbs == 'end':
                s_len = diff_len
                f_len = s_len + org_len
            else:
                s_len = int(diff_len / 2)
                f_len = s_len + org_len

            end_value = np.mean(mote[-gibbs_range:])
            start_value = np.mean(mote[:gibbs_range])
            temp = start_value * np.ones(new_len)
            temp[f_len:] = end_value
            temp[s_len:f_len] = mote
            mote = temp
        else:
            s_len = 0
            f_len = org_len

        wp = cut_off / nyq
        b, a = ss.butter(filter_order, wp, btype=filter_type)
        mote = ss.filtfilt(b, a, mote)
        # removing extra zeros from gibbs effect
        mote = mote[s_len:f_len]  # TODO: don't use -1

        self._values = mote
        self.clear_cache()

    def remove_average(self, section=-1, verbose=-1):
        """
        Calculates the average and removes it from the record
        """

        if verbose == -1:
            verbose = self.verbose

        average = np.mean(self.values[:section])
        self._values = self.values - average

        if verbose:
            print('removed av.: ', average)
        self.clear_cache()

    def remove_poly(self, poly_fit=0):
        """
        Calculates best fit polynomial and removes it from the record
        """

        x = np.linspace(0, 1.0, self.npts)
        cofs = np.polyfit(x, self.values, poly_fit)
        y_cor = 0 * x
        for co in range(len(cofs)):
            mods = x ** (poly_fit - co)
            y_cor += cofs[co] * mods
        self._values = self.values - y_cor
        self.clear_cache()

    def get_section_average(self, start=0, end=-1, index=False):
        """
        Gets the average value of a part of series.

        Common use is so that it can be patched with another record.

        :param start: int or float, optional, Section start point
        :param end: int or float, optional, Section end point
        :param index: bool, optional, if False then start and end are considered values in time.
        :return float, The mean value of the section.
        """
        return get_section_average(self, start=start, end=end, index=index)

    def add_constant(self, constant):
        """
        Adds a single value from every value in the signal.
        :param constant:
        :return:
        """
        self._values = self.values + constant
        self.clear_cache()

    def add_series(self, series):
        """
        Adds a single value from every value in the signal.
        :param series: A series of values
        :return:
        """
        if len(series) == self.npts:
            self._values = self.values + series
        else:
            raise exceptions.SignalProcessingError("new series has different length to Signal")
        self.clear_cache()

    def add_signal(self, new_signal):
        """
        Combines a signal.
        :param new_signal: Signal object
        :return:
        """
        if isinstance(new_signal, Signal):
            if new_signal.dt == self.dt:
                self.add_series(new_signal.values)
            else:
                raise exceptions.SignalProcessingError("New signal has different time step")
        else:
            raise exceptions.SignalProcessingError("New signal is not a Signal object")

    def running_average(self, width=1):
        """
        Averages the values over a width (chunk of numbers)
        replaces the value in the centre of the chunk.

        :param width: The range over which values are averaged
        """

        mot = self.values

        for i in range(len(mot)):
            if i < width / 2:
                cc = i + int(width / 2) + 1
                self._values[i] = np.mean(mot[:cc])
            elif i > len(mot) - width / 2:
                cc = i - int(width / 2)
                self._values[i] = np.mean(mot[cc:])
            else:
                cc1 = i - int(width / 2)
                cc2 = i + int(width / 2) + 1
                self._values[i] = np.mean(mot[cc1:cc2])

        self.clear_cache()


class AccSignal(Signal):
    _cached_response_spectra = False
    _cached_disp_and_velo = False
    _cached_xi = 0.05
    _s_a = None
    _s_v = None
    _s_d = None
    pga = 0.0
    pgv = 0.0
    pgd = 0.0
    t_b01 = 0.0
    t_b05 = 0.0
    t_b10 = 0.0
    a_rms01 = 0.0
    a_rms05 = 0.0
    a_rms10 = 0.0
    t_595 = 0.0  # significant duration
    sd_start = 0.0  # start time of significant duration
    sd_end = 0.0  # end time of significant duration
    arias_intensity_series = None
    arias_intensity = 0.0
    cav_series = None
    cav = 0.0

    def __init__(self, values, dt, label='m1', smooth_freq_range=(0.1, 30), verbose=0, response_times=(0.1, 5), ccbox=0):
        """
        A time series object
        :param values: An acceleration time series, type=acceleration, should be in m/s2
        :param dt: time step
        """
        super(AccSignal, self).__init__(values, dt, label=label, smooth_freq_range=smooth_freq_range, verbose=verbose, ccbox=ccbox)
        if len(response_times) == 2:
            self.response_times = np.linspace(response_times[0], response_times[1], 100)
        else:
            self.response_times = np.array(response_times)
        self._velocity = np.zeros(self.npts)
        self._displacement = np.zeros(self.npts)

    def clear_cache(self):
        self._cached_smooth_fa = False
        self._cached_fa = False
        self._cached_response_spectra = False
        self._cached_disp_and_velo = False
        self.reset_all_motion_stats()

    def generate_response_spectrum(self, response_times=None, xi=-1):
        """
        Generate the response spectrum for the response_times for a given
        damping (xi). default xi = 0.05
        """
        if self.verbose:
            print('Generating response spectra')
        if response_times is not None:
            self.response_times = response_times

        if xi == -1:
            xi = self._cached_xi
        self._s_d, self._s_v, self._s_a = dh.pseudo_response_spectra(self.values, self.dt, self.response_times, xi)
        self._cached_response_spectra = True

    @property
    def s_a(self):
        """
        Pseudo spectral response acceleration of linear SDOF
        """
        if not self._cached_response_spectra:
            self.generate_response_spectrum()
        return self._s_a

    @property
    def s_v(self):
        """
        Pseudo spectral response velocity of linear SDOF
        """
        if not self._cached_response_spectra:
            self.generate_response_spectrum()
        return self._s_v

    @property
    def s_d(self):
        """
        Spectral response displacement of linear SDOF
        """
        if not self._cached_response_spectra:
            self.generate_response_spectrum()
        return self._s_d

    def correct_me(self):
        """
        This provides a correction to an acceleration time series
        """

        self.remove_average()
        self.butter_pass([0.1, 10])

        disp = ss.detrend(self.displacement)
        vel = np.zeros(self.npts)
        acc = np.zeros(self.npts)
        for i in range(self.npts - 1):  # MEANS THAT ACC has a pulse at the end.
            vel[i + 1] = (disp[i + 1] - disp[i]) / self.dt
            acc[i + 1] = (vel[i + 1] - vel[i]) / self.dt
        self._values = acc
        self.clear_cache()

    def rebase_displacement(self):
        """
        Correction to make the displacement zero at the end of the record
        """

        end_disp = self.displacement[-1]

        acceleration_correction = 2 * end_disp / (self.dt * self.npts)
        self._values -= acceleration_correction
        self.clear_cache()

    def generate_displacement_and_velocity_series(self, trap=True):
        """
        Calculates the displacement and velocity
        """
        self._velocity, self._displacement = sd.velocity_and_displacement_from_acceleration(self.values,
                                                                                          self.dt, trap=trap)
        self._cached_disp_and_velo = True

    @property
    def velocity(self):
        if not self._cached_disp_and_velo:
            self.generate_displacement_and_velocity_series()
        return self._velocity

    @property
    def displacement(self):
        if not self._cached_disp_and_velo:
            self.generate_displacement_and_velocity_series()
        return self._displacement

    def generate_peak_values(self):
        """
        Determines the peak signal values
        """
        self.pga = sm.calculate_peak(self.values)
        self.pgv = sm.calculate_peak(self.velocity)
        self.pgd = sm.calculate_peak(self.displacement)

    def generate_duration_stats(self):
        abs_motion = abs(self.values)

        time = np.arange(self.npts) * self.dt
        # Bracketed duration
        ind01 = np.where(abs_motion / 9.8 > 0.01)  # 0.01g
        time2 = time[ind01]
        try:
            self.t_b01 = time2[-1] - time2[0]
            # rms acceleration in m/s/s
            self.a_rms01 = np.sqrt(1 / self.t_b01 * np.trapz((self.values[ind01[0][0]:ind01[0][-1]]) ** 2, dx=self.dt))
        except IndexError:
            self.t_b01 = -1.
            self.a_rms01 = -1.
        ind05 = np.where(abs_motion / 9.8 > 0.05)  # 0.05g
        time05 = time[ind05]
        try:
            self.t_b05 = time05[-1] - time05[0]
            self.a_rms05 = np.sqrt(1 / self.t_b05 * np.trapz((self.values[ind05[0][0]:ind05[0][-1]]) ** 2, dx=self.dt))
        except IndexError:
            self.t_b05 = -1.
            self.a_rms05 = -1.
        ind10 = np.where(abs_motion / 9.8 > 0.1)  # 0.10g
        time10 = time[ind10]
        try:
            self.t_b10 = time10[-1] - time10[0]
            self.a_rms10 = np.sqrt(1 / self.t_b10 * np.trapz((self.values[ind10[0][0]:ind10[0][-1]]) ** 2, dx=self.dt))
        except IndexError:
            self.t_b10 = -1.
            self.a_rms10 = -1.

        # Trifunac and Brady
        self.sd_start, self.sd_end = sm.significant_duration(self.values, self.dt)

        self.t_595 = self.sd_end - self.sd_start

    def generate_cumulative_stats(self):
        """

        CAV:

        ref:
        Electrical Power Research Institute. Standardization of the Cumulative
        Absolute Velocity. 1991. EPRI TR-100082-1'2, Palo Alto, California.
        """
        # Arias intensity in m/s
        acc2 = self.values ** 2
        self.arias_intensity_series = np.pi / (2 * 9.81) * scipy.integrate.cumtrapz(acc2, dx=self.dt, initial=0)
        self.arias_intensity = self.arias_intensity_series[-1]
        abs_acc = np.abs(self.values)
        self.cav_series = scipy.integrate.cumtrapz(abs_acc, dx=self.dt, initial=0)
        self.cav = self.cav_series[-1]
        # self.arias_intensity = np.pi / (2 * 9.81) * np.trapz(acc2, dx=self.dt)

    def generate_all_motion_stats(self):
        """
        This calculates the duration, Arias intensity
        """

        self.generate_peak_values()
        self.generate_duration_stats()
        self.generate_cumulative_stats()

    def reset_all_motion_stats(self):
        self.pga = 0.0
        self.pgv = 0.0
        self.pgd = 0.0
        self.t_b01 = 0.0
        self.t_b05 = 0.0
        self.t_b10 = 0.0
        self.a_rms01 = 0.0
        self.a_rms05 = 0.0
        self.a_rms10 = 0.0
        self.t_595 = 0.0
        self.sd_start = 0.0
        self.sd_end = 0.0
        self.arias_intensity = 0.0

    # deprecated
    def relative_displacement_response(self, period, xi):
        return dh.single_elastic_response(self.values, self.dt, period, xi)

    def response_series(self, response_times=None, xi=-1):
        """
        Generate the response time series for a set of elastic SDOFs for a given
        damping (xi).
        """
        if self.verbose:
            print('Generating response series')
        if response_times is not None:
            self.response_times = response_times
        if xi == -1:
            xi = self._cached_xi
        resp_u, resp_v, resp_a = dh.response_series(self.values, self.dt, self.response_times, xi)
        return resp_u, resp_v, resp_a


def get_section_average(series, start=0, end=-1, index=False):
    """
    Gets the average value of a part of series.

    Common use is so that it can be patched with another record.

    :param series: A TimeSeries object
    :param start: int or float, optional, Section start point
    :param end: int or float, optional, Section end point
    :param index: bool, optional, if False then start and end are considered values in time.
    :return float, The mean value of the section.
    """
    s_index, e_index = time_indices(series.npts, series.dt, start, end, index)

    section_average = np.mean(series.values[s_index:e_index])
    return section_average


def time_indices(npts, dt, start, end, index):
    """
    Determine the new start and end indices of the time series.

    :param npts: Number of points in original time series
    :param dt: Time step of original time series
    :param start: int or float, optional, New start point
    :param end: int or float, optional, New end point
    :param index: bool, optional, if False then start and end are considered values in time.
    :return: tuple, start index, end index
    """
    if index is False:  # Convert time values into indices
        if end != -1:
            e_index = int(end / dt) + 1
        else:
            e_index = end
        s_index = int(start / dt)
    else:
        s_index = start
        e_index = end
    if e_index > npts:
        raise exceptions.SignalProcessingWarning("Cut point is greater than time series length")
    return s_index, e_index


__author__ = 'maximmillen'

import matplotlib.pyplot as plt
import numpy as np
import engformat.plot as esfp

from eqsig.exceptions import SignalProcessingError

# from bwplot import spectra as cbox
from bwplot import cbox


def plot_response_spectrum(rec, **kwargs):
    rec.generate_response_spectrum()
    plot_on = 1
    legend_off = kwargs.get('legend_off', False)
    ratio = kwargs.get('ratio', False)
    info_str = kwargs.get('info_str', '')

    label = kwargs.get('label', rec.label)

    time = kwargs.get('time', "frequency")
    sub_plot = kwargs.get('sub_plot', 0)
    response_type = kwargs.get('type', 'acceleration')
    colour = kwargs.get('colour', False)
    ccbox = kwargs.get('ccbox', "auto")
    scale = kwargs.get('scale', 1.0)
    log_off = kwargs.get('log_off', False)
    verbose = kwargs.get('verbose', 0)
    title = kwargs.get('title', False)

    if response_type == 'acceleration':
        response = rec.s_a
        y_label = 'Spectral acceleration [$m/s^2$]'
    elif response_type == 'velocity':
        response = rec.s_v
        y_label = 'Spectral velocity [$m/s$]'
    elif response_type == 'displacement':
        response = rec.s_d
        y_label = 'Spectral displacement [$m$]'
    else:
        raise NotImplementedError
    if sub_plot == 0:
        sub_plot = plt.figure().add_subplot(111)
    else:
        plot_on = 0
    if time == "period":
        x_para = rec.response_times
    else:
        x_para = 1 / rec.response_times

    if ccbox == "auto":
            ccbox = len(sub_plot.lines)
    rec.ccbox = ccbox

    sub_plot.plot(x_para, response * scale, label=label, c=cbox(0 + rec.ccbox), lw=0.7)

    if not log_off:
        sub_plot.set_xscale('log')

    if time == "period":
        if verbose:
            print('setting x-axis as period')
        sub_plot.set_xlabel('Time period [s]')
        sub_plot.set_xlim([0, rec.response_times[-1]])
    else:
        sub_plot.set_xscale('log')
        sub_plot.set_xlabel('Frequency [Hz]')
    sub_plot.set_ylabel(y_label)

    if title is not False:
        sub_plot.set_title(title)

    if legend_off is False:
        sub_plot.legend(loc='upper left', prop={'size': 8})
    if plot_on == 1:

        plt.show()
    else:
        return sub_plot


def plot_time_series(rec, **kwargs):
    plot_on = kwargs.get('plot_on', False)
    legend_off = kwargs.get('legend_off', False)
    info_str = kwargs.get('info_str', '')
    motion_type = kwargs.get('motion_type', 'acceleration')

    y_label = kwargs.get('y_label', motion_type.title())
    x_label = kwargs.get('x_label', 'Time [s]')
    y_limits = kwargs.get('y_limits', False)
    label = kwargs.get('label', rec.label)
    sub_plot = kwargs.get('sub_plot', 0)
    window = kwargs.get('window', [0, -1])
    ccbox = kwargs.get('ccbox', "auto")  # else integer
    
    cut_index = np.array([0, len(rec.motion)])
    if window[0] != 0:
        cut_index[0] = int(window[0] / rec.dt)
    if window[1] != -1:
        cut_index[1] = int(window[1] / rec.dt)

    
    if sub_plot == 0:
        sub_plot = plt.figure().add_subplot(111)
    else:
        plot_on = 0
    if ccbox == "auto":
        ccbox = len(sub_plot.lines)

    if motion_type == "acceleration":
        motion = rec.motion[cut_index[0]:cut_index[1]]
        balance = True
    elif motion_type == "velocity":
        rec.calculate_displacements()
        motion = rec.velocity[cut_index[0]:cut_index[1]]
        balance = True
    elif motion_type == "displacement":
        rec.calculate_displacements()
        motion = rec.displacement[cut_index[0]:cut_index[1]]
        balance = False
    elif motion_type == "custom":
        motion = rec.motion[cut_index[0]:cut_index[1]]
        balance = False
    else:
        raise NotImplementedError

    t0 = rec.dt * (np.linspace(0, (len(motion) + 1), (len(motion))) + cut_index[0])
    sub_plot.plot(t0, motion, label=label, c=cbox(0 + ccbox), lw=0.7)
    # sub_plot.plot(rec.dt * cut_index, [0, 0], c="k", lw=0.5, zorder=0)  # TODO: use plot-tricks
    # esfp.time_series(sub_plot)
    esfp.time_series(sub_plot, balance=balance)

    if x_label is not False:
        sub_plot.set_xlabel(x_label)
    sub_plot.set_ylabel(y_label)
    if info_str != '':
        stitle = 'Time series \n' + info_str
        sub_plot.set_title(stitle)
    if y_limits is not False:
        sub_plot.set_ylim(y_limits)

    x_limits = [0, t0[-1]]
    sub_plot.set_xlim(x_limits)
    if not legend_off:
        sub_plot.legend(loc='upper right', prop={'size': 8})
    if plot_on == 1:
        plt.show()
    else:
        return sub_plot


def plot_avd(rec, sub_plots=None, ccbox=0, **kwargs):
    """
    Plot acceleration, velocity and displacement
    :return:
    """
    label = kwargs.get('label', rec.label)
    legend_off = kwargs.get('legend_off', False)
    if sub_plots is None:
        big_fig = plt.figure()
        sub_plots = [big_fig.add_subplot(311),
                    big_fig.add_subplot(312),
                    big_fig.add_subplot(313)]
    plot_time_series(rec, ccbox=ccbox, sub_plot=sub_plots[0], label=label, legend_off=True)
    plot_time_series(rec, ccbox=ccbox, sub_plot=sub_plots[1], motion_type="velocity", legend_off=True)
    plot_time_series(rec, ccbox=ccbox, sub_plot=sub_plots[2], motion_type="displacement", legend_off=True)

    if not legend_off:
        sub_plots[0].legend()
    return sub_plots

    
def plot_fa_spectrum(rec, **kwargs):
    plot_on = kwargs.get('plot_on', False)
    legend_off = kwargs.get('legend_off', False)
    smooth = kwargs.get('smooth', False)

    info_str = kwargs.get('info_str', '')

    label = kwargs.get('label', rec.label)

    log_off = kwargs.get('log_off', False)
    title = kwargs.get('title', 'Fourier amplitude acceleration spectrums \n' + info_str)

    band = kwargs.get('band', 40)
    sub_plot = kwargs.get('sub_plot', 0)
    ccbox = kwargs.get('ccbox', 'auto')


    if smooth is False:
        spectrum = abs(rec.fa_spectrum)
        frequencies = rec.fa_frequencies

    else:

        rec.generate_smooth_fa_spectrum(band=band)
        spectrum = abs(rec.smooth_fa_spectrum)
        frequencies = rec.smooth_fa_frequencies


    if sub_plot == 0:
        sub_plot = plt.figure().add_subplot(111)
    else:
        plot_on = 0
    if ccbox == "auto":
        ccbox = len(sub_plot.lines)
    rec.ccbox = ccbox

    sub_plot.plot(frequencies, spectrum, label=label, c=cbox(ccbox), lw=0.7)

    sub_plot.set_xscale('log')
    if log_off is not True:
        sub_plot.set_yscale('log')
    sub_plot.set_xlabel('Frequency [Hz]')
    sub_plot.set_ylabel('Fourier Amplitude [m/s2]')

    if title is not False:
        sub_plot.set_title(title)
    if legend_off is False:
        sub_plot.legend(loc='upper left', prop={'size': 8})
    if plot_on == 1:
        plt.show()
    else:
        return sub_plot


def plot_transfer_function(base_rec, recs, **kwargs):
    """
    Plots the transfer function between the base values and a list of other motions
    :param base_rec: A Record object
    :param recs: A list of Record objects
    :param kwargs:
    :return:
    """
    plot_on = kwargs.get('plot_on', False)
    legend_off = kwargs.get('legend_off', False)
    smooth = kwargs.get('smooth', False)

    info_str = kwargs.get('info_str', '')

    base_label = kwargs.get('label', base_rec.label)

    log_off = kwargs.get('log_off', False)
    title = kwargs.get('title', 'Transfer function \n' + info_str)

    band = kwargs.get('band', 40)
    sub_plot = kwargs.get('sub_plot', 0)
    ccbox = kwargs.get('ccbox', 'auto')

    if smooth is False:
        base_spectrum = abs(base_rec.fa_spectrum)
        base_frequencies = base_rec.fa_frequencies
    else:
        base_rec.generate_smooth_fa_spectrum(band=band)
        base_spectrum = abs(base_rec.smooth_fa_spectrum)
        base_frequencies = base_rec.smooth_fa_frequencies


    if sub_plot == 0:
        sub_plot = plt.figure().add_subplot(111)
    else:
        plot_on = 0
    if ccbox == "auto":
        ccbox = len(sub_plot.lines)

    for rec in recs:
        if smooth is False:
            if rec.dt != base_rec.dt or len(rec.motion) != len(base_rec.motion):
                raise SignalProcessingError("Motion lengths and timestep do not match. "
                                            "Cannot build non-smooth spectrum")
            spectrum = abs(rec.fa_spectrum)
        else:
            rec.reset_smooth_spectrum(freq_range=base_rec.freq_range)
            rec.generate_smooth_fa_spectrum(band=band)
            spectrum = abs(rec.smooth_fa_spectrum)

        sub_plot.plot(base_frequencies, spectrum / base_spectrum, label=rec.label, c=cbox(ccbox), lw=0.7)

    sub_plot.set_xscale('log')
    if log_off is not True:
        sub_plot.set_yscale('log')
    sub_plot.set_xlabel('Frequency [Hz]')
    sub_plot.set_ylabel('Transfer function from %s' % base_label)

    if title is not False:
        sub_plot.set_title(title)
    if legend_off is False:
        sub_plot.legend(loc='upper left', prop={'size': 8})
    if plot_on == 1:
        plt.show()
    else:
        return sub_plot